




#include "CliArguments.h"


#include "bitgenConfig.h"
#include "OtpFolderHandler.h"



CliArguments CliArguments::argcvToCliArgument(int argc, char* argv[])
{
	std::vector<std::string> rawArgs;
	for(int i = 1 ; i < argc ; i++)
	{
		rawArgs.push_back(argv[i]);		
	}

	const CliArguments args(rawArgs);
	return args;
}



CliArguments::CliArguments(const std::vector<std::string>& inArgs, const int startAt) : args(inArgs), startPos(startAt) {}



std::string CliArguments::prefixString() const
{
	int num(0);

	std::string ret;
	for(const std::string& str : args)
	{		
		num++;

		if(num > startPos)
		{
			break;
		}		
		
		ret += str;
		ret += " ";


	}

	return ret;
}



std::string CliArguments::toString() const
{
	int num(0);

	std::string ret;
	for(const std::string& str : args)
	{		
		num++;
		if(num <= startPos)
		{
			continue;
		}		
		
		ret += str;
		ret += " ";
	}
	return ret;
}



std::string CliArguments::operator[](const int i) const
{
	const int pos = indexPos(i); //i + startPos;
	if(pos >= args.size())
	{
		throw std::string("Argument read out of bounds");
	}
		
	return args[pos];
}


	
	
std::vector<std::string> CliArguments::getExtraArgs(const int myStart) const
{
	const int start = myStart + startPos;
	std::vector<std::string> ret;
	int num(0);
	
	for(std::vector<std::string>::const_iterator it = args.begin() ;
		it != args.end() ;
		++it)
	{
		const std::string& str = (*it);
		if(num >= start)
		{
			ret.push_back(str);
		}
		
		num++;
	}
	return ret;
}




GenerationParameters CliArguments::parseTheExtraParams(const int startPos, const bool useCurrencySpecific) const
{
	//std::cout << "In parseTheExtraParams" << std::endl;
	GenerationParameters params;	

	const std::vector<std::string> allExtraArgs = getExtraArgs(startPos);

	ExtraOptionsRandom extraOptions;
	
		
	for(std::vector<std::string>::const_iterator it = allExtraArgs.begin() ;
		it != allExtraArgs.end() ;
		++it)
	{
		const std::string& arg = (*it);
		
		if(params.useArg(arg))
		{
			//Nothing
		}
		else if(extraOptions.useArg(arg))
		{
			//Nothing			
		}
		else if(useCurrencySpecific && params.useCurrencySpecific(arg))
		{
			//Nothing			
		}		
		else
		{
			std::cout << "Incorrect argument: " << arg << std::endl;
			throw std::string("Argument error");			
		}
	}
	
	params.checkNameProvided();
	return params;
}



std::string parse256BitHexString(const std::string& hexVal)
{
	const int numBytes = 32;
	
	if(hexVal.size() != (numBytes * 2))
	{
		std::cerr << std::endl;
		std::cerr << "Incorrect amount of data given. Got " << hexVal.size() << " hex digits, want " << (numBytes * 2) << " hex digits" << std::endl;
		throw std::string("Not enough data");
	}
		
	const std::string binStr = convertFromHex(hexVal);
	return binStr;
}





std::list<BigInt<256> > CliArguments::getPadsFromCommandline(bool& useStaticXorIfPresent) const
{
	std::list<BigInt<256> > pads;
	
	const int numArgs = args.size(); // - 1;
	
	for(int i = startPos ; i < numArgs ; i++)
	{
		const std::string padHex = args[i];
		
		if(padHex == "nostatic")
		{
			useStaticXorIfPresent = false;
			return pads;
		}
	
		const std::string padBin = parse256BitHexString(padHex);	
	
		const BigInt<256> padInt(padBin,'c');
		//std::cout << "FOUND PAD: " << hexString(padInt.getBinaryLong()) << std::endl;
		pads.push_back(padInt);
	}

	return pads;
}




std::list<BigInt<256> > getPadsWithNum(const int addressNum, std::string& genTime)
{
	std::list<BigInt<256>> ret;
	
	std::string publicFolder;
	const std::vector<std::string> folders = BitgenConfig::getPublicAndOtpFoldersVector(publicFolder);	
	
	for(int i = 0 ; i < folders.size() ; i++)
	{
		const std::string folder = folders[i];		
		
		const int padOneIndex = i + 1;
		
		const std::string totFolder = OtpFolderHandler::getTotalPrivFolder(folder, padOneIndex, addressNum);		
		const EntryInfo info = OtpFolderHandler::readEntryInfo(totFolder, addressNum, padOneIndex);
		
		if(i == 0)
		{
			genTime = info.date;
		}
		
		const std::string padBin = parse256BitHexString(info.pad);		
		const BigInt<256> padInt(padBin,'c');				
		ret.push_back(padInt);
	}

	return ret;
}



std::list<BigInt<256> > CliArguments::getPadsFromArgs(bool& useStaticXorIfPresent, GenerationParameters& params, int& addressNum, std::string& genTime) const
{
	useStaticXorIfPresent = true;

	if(size() == 1)
	{
		
		const std::string addressNumStr = args[1];
		addressNum = atoi(addressNumStr.c_str());
		
		std::string accountText("From OTP ");
		accountText += intToString(addressNum);
		accountText += " do not use";
				
		params.setAccountName(accountText);
		
		const std::list<BigInt<256> > padList = getPadsWithNum(addressNum, genTime);		
		return padList;
	}
	else
	{
		if(args.size() < 2)
		{
			std::cout << "Usage: bitgen otp [OTP1 ... OTPn | OTPNUM]" << std::endl;
			throw std::string("Incorrect usage");
		}

		params.setAccountName("From OTP do not use");
		
		const std::list<BigInt<256> > padList = getPadsFromCommandline(useStaticXorIfPresent);
		return padList;
	}
}


